import datetime
import inspect
from typing import Union
from functools import wraps

from starlette.responses import StreamingResponse
from logdetective.server.database.models import EndpointType, AnalyzeRequestMetrics
from logdetective.server import models


def add_new_metrics(
    api_name: str, build_log: models.BuildLog, received_at: datetime.datetime = None
) -> int:
    """Add a new database entry for a received request.

    This will store the time when this function is called,
    the endpoint from where the request was received,
    and the log for which analysis is requested.
    """
    return AnalyzeRequestMetrics.create(
        endpoint=EndpointType(api_name),
        log_url=build_log.url,
        request_received_at=received_at
        if received_at
        else datetime.datetime.now(datetime.timezone.utc),
    )


def update_metrics(
    metrics_id: int,
    response: Union[models.Response, models.StagedResponse, StreamingResponse],
    sent_at: datetime.datetime = None,
) -> None:
    """Update a database metric entry for a received request,
    filling data for the given response.

    This will add to the database entry the time when the response was sent,
    the length of the created response and the certainty for it.
    """
    response_sent_at = (
        sent_at if sent_at else datetime.datetime.now(datetime.timezone.utc)
    )
    response_length = None
    if hasattr(response, "explanation") and "choices" in response.explanation:
        response_length = sum(
            len(choice["text"])
            for choice in response.explanation["choices"]
            if "text" in choice
        )
    response_certainty = (
        response.response_certainty if hasattr(response, "response_certainty") else None
    )
    AnalyzeRequestMetrics.update(
        metrics_id, response_sent_at, response_length, response_certainty
    )


def track_request():
    """
    Decorator to track requests metrics
    """

    def decorator(f):
        @wraps(f)
        async def async_decorated_function(*args, **kwargs):
            metrics_id = add_new_metrics(f.__name__, kwargs["build_log"])
            response = await f(*args, **kwargs)
            update_metrics(metrics_id, response)
            return response

        @wraps(f)
        def sync_decorated_function(*args, **kwargs):
            metrics_id = add_new_metrics(f.__name__, kwargs["build_log"])
            response = f(*args, **kwargs)
            update_metrics(metrics_id, response)
            return response

        if inspect.iscoroutinefunction(f):
            return async_decorated_function
        return sync_decorated_function

    return decorator



import enum
import datetime

from typing import Optional
from sqlalchemy import Column, Integer, Float, DateTime, String, Enum

from logdetective.server.database.base import Base, transaction


class EndpointType(enum.Enum):
    """Different analyze endpoints"""

    ANALYZE = "analyze_log"
    ANALYZE_STAGED = "analyze_log_staged"
    ANALYZE_STREAM = "analyze_log_stream"


class AnalyzeRequestMetrics(Base):
    """Store data related to received requests and given responses"""

    __tablename__ = "analyze_request_metrics"

    id = Column(Integer, primary_key=True)
    endpoint = Column(
        Enum(EndpointType),
        nullable=False,
        index=True,
        comment="The service endpoint that was called",
    )
    request_received_at = Column(
        DateTime,
        nullable=False,
        index=True,
        default=datetime.datetime.now(datetime.timezone.utc),
        comment="Timestamp when the request was received",
    )
    log_url = Column(
        String,
        nullable=False,
        index=False,
        comment="Log url for which analysis was requested",
    )
    response_sent_at = Column(
        DateTime, nullable=True, comment="Timestamp when the response was sent back"
    )
    response_length = Column(
        Integer, nullable=True, comment="Length of the response in chars"
    )
    response_certainty = Column(
        Float, nullable=True, comment="Certainty for generated response"
    )

    @classmethod
    def create(
        cls,
        endpoint: EndpointType,
        log_url: str,
        request_received_at: Optional[datetime.datetime] = None,
    ) -> int:
        """Create AnalyzeRequestMetrics new line
        with data related to a received request"""
        with transaction(commit=True) as session:
            metrics = AnalyzeRequestMetrics()
            metrics.endpoint = endpoint
            metrics.request_received_at = request_received_at or datetime.datetime.now(
                datetime.timezone.utc
            )
            metrics.log_url = log_url
            session.add(metrics)
            session.flush()
            return metrics.id

    @classmethod
    def update(
        cls,
        id_: int,
        response_sent_at: datetime,
        response_length: int,
        response_certainty: float,
    ) -> None:
        """Update an AnalyzeRequestMetrics line
        with data related to the given response"""
        with transaction(commit=True) as session:
            metrics = session.query(AnalyzeRequestMetrics).filter_by(id=id_).first()
            metrics.response_sent_at = response_sent_at
            metrics.response_length = response_length
            metrics.response_certainty = response_certainty
            session.add(metrics)



# pylint: disable=line-too-long
DEFAULT_ADVISOR = "fedora-copr/Mistral-7B-Instruct-v0.2-GGUF"

PROMPT_TEMPLATE = """
Given following log snippets, and nothing else, explain what failure, if any, occured during build of this package.

Analysis of the snippets must be in a format of [X] : [Y], where [X] is a log snippet, and [Y] is the explanation.
Snippets themselves must not be altered in any way whatsoever.

Snippets are delimited with '================'.

Finally, drawing on information from all snippets, provide complete explanation of the issue and recommend solution.

Snippets:

{}

Analysis:

"""

SUMMARIZE_PROMPT_TEMPLATE = """
Does following log contain error or issue?

Log:

{}

Answer:

"""

SNIPPET_PROMPT_TEMPLATE = """
Analyse following RPM build log snippet. Describe contents accurately, without speculation or suggestions for resolution.

Snippet:

{}

Analysis:

"""

PROMPT_TEMPLATE_STAGED = """
Given following log snippets, their explanation, and nothing else, explain what failure, if any, occured during build of this package.

Snippets are in a format of [X] : [Y], where [X] is a log snippet, and [Y] is the explanation.

Snippets are delimited with '================'.

Drawing on information from all snippets, provide complete explanation of the issue and recommend solution.

Snippets:

{}

Analysis:

"""

SNIPPET_DELIMITER = "================"
